@@ -143,6 +143,14 @@
 
 ;;;; Storage allocation:
 
+(defun generate-stack-overflow-check (vop size)
+  (let ((overflow (generate-error-code vop
+                                       'stack-allocated-object-overflows-stack-error
+                                       size)))
+    (loadw tmp-tn thread-tn thread-control-stack-end-slot)
+    (inst sub tmp-tn tmp-tn csp-tn)
+    (inst cmp tmp-tn size)
+    (inst b :le overflow)))
 
 ;;; This is the main mechanism for allocating memory in the lisp heap.
 ;;;
@@ -158,12 +166,14 @@
 ;;; surround a call to ALLOCATION anyway), and to indicate that the
 ;;; P-A FLAG-TN is also acceptable here.
 
-#+gencgc
-(defun allocation-tramp (type alloc-tn size back-label)
+#+generational
+(defun allocation-tramp (type alloc-tn size back-label systemp)
   (if (integerp size)
       (load-immediate-word tmp-tn size)
       (inst mov tmp-tn size))
-  (let ((asm-routine (if (eq type 'list) 'list-alloc-tramp 'alloc-tramp)))
+  (let ((asm-routine (if systemp
+                         (if (eq type 'list) 'sys-list-alloc-tramp 'sys-alloc-tramp)
+                         (if (eq type 'list) 'list-alloc-tramp 'alloc-tramp))))
     (invoke-asm-routine asm-routine alloc-tn))
   (inst b back-label))
 
@@ -172,17 +182,25 @@
 (defun allocation (type size lowtag result-tn
                    &key flag-tn
                         stack-allocate-p
-                        overflow)
+                        overflow
+                        (systemp (system-tlab-p type (sb-c::vop-node sb-assem::*current-vop*))))
   (declare (ignorable type))
   ;; Normal allocation to the heap.
   (if stack-allocate-p
       (assemble ()
-        (inst add tmp-tn csp-tn lowtag-mask)
-        (inst and tmp-tn tmp-tn (lognot lowtag-mask))
-        (inst add csp-tn tmp-tn (add-sub-immediate size result-tn))
-        (inst add result-tn tmp-tn lowtag))
+        (cond ((aligned-stack-p stack-allocate-p)
+               (move tmp-tn csp-tn)
+               (inst add csp-tn csp-tn (add-sub-immediate size result-tn))
+               (inst add result-tn tmp-tn lowtag))
+              (t
+               (inst add tmp-tn csp-tn lowtag-mask)
+               (inst and tmp-tn tmp-tn (lognot lowtag-mask))
+               (inst add csp-tn tmp-tn (add-sub-immediate size result-tn))
+               (inst add result-tn tmp-tn lowtag))))
       (let ((alloc (gen-label))
-            #+sb-thread (tlab (if (eq type 'list) thread-cons-tlab-slot thread-mixed-tlab-slot))
+            #+sb-thread (tlab (if systemp
+                                  (if (eq type 'list) thread-sys-cons-tlab-slot thread-sys-mixed-tlab-slot)
+                                  (if (eq type 'list) thread-cons-tlab-slot thread-mixed-tlab-slot)))
             #-sb-thread (region-offset (if (eq type 'list)
                                            cons-region-offset
                                            mixed-region-offset))
@@ -208,7 +226,8 @@
               (allocation-tramp type
                                 result-tn
                                 size
-                                BACK-FROM-ALLOC))))))
+                                BACK-FROM-ALLOC
+                                systemp))))))
 
 (defmacro with-fixed-allocation ((result-tn flag-tn type-code size
                                             &key (lowtag other-pointer-lowtag)
@@ -378,19 +397,30 @@
         (:temporary (:sc non-descriptor-reg) ea)
         (:vop-var vop)
         (:generator 2
-          (sc-case index
-            (immediate
-             (inst add ea object (load-store-offset
-                                  (- (ash (+ ,offset (tn-value index)) word-shift)
-                                     ,lowtag))))
-            (t
-             ;; Scale the index
-             (let ((unshift (if (sc-is index any-reg) n-fixnum-tag-bits 0)))
-               (inst add ea object (lsl index (- word-shift unshift))))
-             ;; Calculate the exact cell address to ensure the right card is marked
-             (inst add ea ea (- (ash vector-data-offset word-shift) other-pointer-lowtag))))
-          (emit-gengc-barrier object ea tmp-tn (vop-nth-arg 2 vop) value)
-          (storew value ea 0 0))))
+          (cond ((require-gengc-barrier-p object (vop-nth-arg 2 vop) value)
+                 (sc-case index
+                   (immediate
+                    (inst add ea object (load-store-offset
+                                         (- (ash (+ ,offset (tn-value index)) word-shift)
+                                            ,lowtag))))
+                   (t
+                    ;; Scale the index
+                    (let ((unshift (if (sc-is index any-reg) n-fixnum-tag-bits 0)))
+                      (inst add ea object (lsl index (- word-shift unshift))))
+                    ;; Calculate the exact cell address to ensure the right card is marked
+                    (inst add ea ea (- (ash vector-data-offset word-shift) other-pointer-lowtag))))
+                 (emit-gengc-barrier object ea tmp-tn t)
+                 (storew value ea 0 0))
+                (t
+                 (sc-case index
+                   (immediate
+                    (inst str value (@ object (load-store-offset
+                                               (- (ash (+ ,offset (tn-value index)) word-shift)
+                                                  ,lowtag)))))
+                   (t
+                    (let ((unshift (if (sc-is index any-reg) n-fixnum-tag-bits 0)))
+                      (inst add tmp-tn object (lsl index (- word-shift unshift))))
+                    (storew value tmp-tn ,offset ,lowtag))))))))
     (t
      `(define-vop (,name)
         (:translate ,translate)
@@ -400,8 +430,12 @@
                (value :scs (,@scs zero)))
         (:arg-types ,type tagged-num ,el-type)
         (:vop-var vop)
+        ,@(if barrierp '((:gc-barrier 0 2)
+                         (:info barrier)))
         (:generator 2
-          ,@(if barrierp '((emit-gengc-barrier object nil tmp-tn (vop-nth-arg 2 vop) value)))
+          ,@(when barrierp
+              '((when barrier
+                  (emit-gengc-barrier object nil tmp-tn t))))
           (sc-case index
             (immediate
              (inst str value (@ object (load-store-offset
@@ -534,3 +568,31 @@
   `(inst str ,reg (@ thread-tn ,(info :variable :wired-tls symbol)))
   #-sb-thread
   `(store-symbol-value ,reg ,symbol))
+
+;;; Load constants, stack-values, reusing when possible
+(defmacro maybe-load (tn &optional (temp 'temp))
+  (once-only ((tn tn))
+    `(sc-case ,tn
+       ((any-reg descriptor-reg)
+        ,tn)
+       ((immediate constant)
+        (cond ((not (sb-c::tn-leaf ,tn))
+               (load-constant vop ,tn ,temp)
+               ,temp)
+              ((eql (tn-value ,tn) 0)
+               zr-tn)
+              ((or (eql prev-constant (tn-value ,tn))
+                   (progn
+                     (setf prev-constant (tn-value ,tn))
+                     nil))
+               temp)
+              ((sc-is ,tn constant)
+               (load-constant vop ,tn ,temp)
+               ,temp)
+              (t
+               (load-immediate vop ,tn ,temp)
+               ,temp)))
+       (control-stack
+        (setf prev-constant nil)
+        (load-stack-tn ,temp ,tn)
+        ,temp))))
