@@ -29,7 +29,7 @@
 (defun environment-analyze (component)
   (declare (type component component))
   (aver (every (lambda (x)
-                 (eq (functional-kind x) :deleted))
+                 (functional-kind-eq x deleted))
                (component-new-functionals component)))
   (setf (component-new-functionals component) ())
   (dolist (fun (component-lambdas component))
@@ -40,13 +40,13 @@
   (find-non-local-exits component)
   ;; Close over closures.
   (dolist (fun (component-lambdas component))
-    (when (and (eq (functional-kind fun) :external)
+    (when (and (functional-kind-eq fun external)
                (environment-closure (lambda-environment fun)))
       (let ((enclose-env (get-node-environment (xep-enclose fun))))
         (dolist (ref (leaf-refs fun))
           (close-over fun (get-node-environment ref) enclose-env)))))
 
-  (find-dynamic-extent-lvars component)
+  (find-lvar-dynamic-extents component)
   (find-cleanup-points component)
   (tail-annotate component)
   (determine-lambda-var-and-nlx-extent component)
@@ -54,10 +54,10 @@
   (dolist (fun (component-lambdas component))
     (when (null (leaf-refs fun))
       (let ((kind (functional-kind fun)))
-        (unless (or (eq kind :toplevel)
+        (unless (or (eql kind (functional-kind-attributes toplevel))
                     (functional-has-external-references-p fun))
-          (aver (member kind '(:optional :cleanup :escape)))
-          (setf (functional-kind fun) nil)
+          (aver (logtest kind (functional-kind-attributes optional cleanup escape)))
+          (setf (functional-kind fun) (functional-kind-attributes nil))
           (delete-functional fun)))))
 
   (values))
@@ -65,10 +65,11 @@
 ;;; If FUN has an environment, return it, otherwise assign an empty
 ;;; one and return that.
 (defun get-lambda-environment (fun)
-  (declare (type clambda fun))
+  (declare (type clambda fun)
+           (inline make-environment))
   (let ((fun (lambda-home fun)))
     (or (lambda-environment fun)
-        (let ((res (make-environment :lambda fun)))
+        (let ((res (make-environment fun)))
           (setf (lambda-environment fun) res)
           (dolist (lambda (lambda-lets fun))
             (setf (lambda-environment lambda) res))
@@ -151,7 +152,7 @@
 ;;; CLAMBDAs that need checking.
 (defun determine-lambda-var-and-nlx-extent (component)
   (dolist (fun (component-lambdas component))
-    (when (and (eq (functional-kind fun) :external)
+    (when (and (functional-kind-eq fun external)
                ;; We treat DYNAMIC-EXTENT declarations on functions as
                ;; trusted assertions that none of the values closed
                ;; over survive the extent of the function.
@@ -198,7 +199,8 @@
 ;;; actually exiting the scope (i.e. a BLOCK), and will also do any
 ;;; other cleanups that may have to be done on the way.
 (defun insert-nlx-entry-stub (exit env)
-  (declare (type environment env) (type exit exit))
+  (declare (type environment env) (type exit exit)
+           (inline make-nlx-info))
   (let* ((exit-block (node-block exit))
          (next-block (first (block-succ exit-block)))
          (entry (exit-entry exit))
@@ -255,7 +257,7 @@
            (setq info (exit-nlx-info exit))
            (aver info)))
     (close-over info (node-environment exit) env)
-    (when (eq (functional-kind exit-fun) :escape)
+    (when (functional-kind-eq exit-fun escape)
       (mapc (lambda (x)
               (setf (node-derived-type x) *wild-type*))
             (leaf-refs exit-fun))
@@ -283,10 +285,11 @@
 
 ;;; For each downward funarg, mark the funarg as dynamic extent. For
 ;;; now this only works on globally named functions.
-(defun dxify-downward-funargs (node dxable-args fun-name)
-  #+sb-xc-host
-  (declare (ignore fun-name))
-  (let (cleanup)
+(defun dxify-downward-funargs (node)
+  (let* ((fun-name (combination-fun-source-name node nil))
+         (dxable-args (and fun-name
+                           (fun-name-dx-args fun-name)))
+         dynamic-extent)
     ;; Experience shows that users place incorrect DYNAMIC-EXTENT declarations
     ;; without due consideration and care. Since the declaration was ignored
     ;; in more contexts than not, it was relatively harmless.
@@ -297,129 +300,228 @@
     ;; that callers should always use more stack space. You should really
     ;; only do that if you don't also need an arbitrarily long call chain.
     ;; MAP and friends are good examples where this pertains]
-    (when #+sb-xc-host t                ; always trust our own code
-          #-sb-xc-host
-          (or (let ((pkg (sb-xc:symbol-package (fun-name-block-name fun-name))))
-                ;; callee "probably" won't get redefined
-                (or (not pkg)
-                    (package-locked-p pkg)
-                    (system-package-p pkg)
-                    (eq pkg (find-package "COMMON-LISP"))
-                    (basic-combination-fun-info node)))
-              (policy node (= safety 0)))
-          (dolist (arg-spec dxable-args)
-            (when (symbolp arg-spec)
-              ;; If there are keywords, we had better have a FUN-TYPE
-              (let ((fun-type (lvar-type (combination-fun node))))
-                ;; Can't do anything unless we can ascertain where
-                ;; the keyword arguments start.
-                (when (fun-type-p fun-type)
-                  (let* ((keys-index
-                           (+ (length (fun-type-required fun-type))
-                              (length (fun-type-optional fun-type))))
-                         (keywords-supplied
-                           (nthcdr keys-index (combination-args node))))
-                    ;; Everything in a keyword position needs to be
-                    ;; constant.
-                    (loop
-                      (unless (cdr keywords-supplied) (return))
-                      (let ((keyword (car keywords-supplied)))
-                        (unless (constant-lvar-p keyword)
-                          (return))
-                        (when (eq (lvar-value keyword) arg-spec)
-                          ;; Map it to a positional arg
-                          (setq arg-spec (1+ keys-index))
-                          (return))
-                        (setq keywords-supplied (cddr keywords-supplied))
-                        (incf keys-index 2)))))))
-            (when (integerp arg-spec)
-              (let* ((arg (or (nth arg-spec (combination-args node))
-                              (return-from dxify-downward-funargs)))
-                     (use (principal-lvar-use arg)))
-                (when (and (not (lvar-dynamic-extent arg))
-                           ;; We check that the use is a lambda so
-                           ;; that we don't end up getting notes about
-                           ;; not being able to allocate later.
-                           (ref-p use)
-                           (lambda-p (ref-leaf use))
-                           (not (leaf-dynamic-extent (functional-entry-fun (ref-leaf use)))))
-                  (let ((enclose (xep-enclose (ref-leaf use))))
-                    (cond ((enclose-cleanup enclose)
-                           (setf (leaf-dynamic-extent (functional-entry-fun (ref-leaf use)))
-                                 'dynamic-extent))
-                          (t
-                           (unless cleanup
-                             (setq cleanup (insert-dynamic-extent-cleanup node)))
-                           (let ((dx-info (make-dx-info :kind 'dynamic-extent
-                                                        :value arg :cleanup cleanup)))
-                             (setf (lvar-dynamic-extent arg) dx-info)
-                             (push dx-info (cleanup-nlx-info cleanup)))))))))))))
+    (when (and dxable-args
+               #-sb-xc-host                   ; always trust our own code
+               (or (let ((pkg (sb-xc:symbol-package
+                               (fun-name-block-name fun-name))))
+                     ;; callee "probably" won't get redefined
+                     (or (not pkg)
+                         (package-locked-p pkg)
+                         (system-package-p pkg)
+                         (eq pkg *cl-package*)
+                         (basic-combination-fun-info node)))
+                   (policy node (= safety 0))))
+      (dolist (arg-spec dxable-args)
+        (when (symbolp arg-spec)
+          ;; If there are keywords, we had better have a FUN-TYPE
+          (let ((fun-type (lvar-type (combination-fun node))))
+            ;; Can't do anything unless we can ascertain where
+            ;; the keyword arguments start.
+            (when (fun-type-p fun-type)
+              (let* ((keys-index
+                       (+ (length (fun-type-required fun-type))
+                          (length (fun-type-optional fun-type))))
+                     (keywords-supplied
+                       (nthcdr keys-index (combination-args node))))
+                ;; Everything in a keyword position needs to be
+                ;; constant.
+                (loop
+                  (unless (cdr keywords-supplied) (return))
+                  (let ((keyword (car keywords-supplied)))
+                    (unless (constant-lvar-p keyword)
+                      (return))
+                    (when (eq (lvar-value keyword) arg-spec)
+                      ;; Map it to a positional arg
+                      (setq arg-spec (1+ keys-index))
+                      (return))
+                    (setq keywords-supplied (cddr keywords-supplied))
+                    (incf keys-index 2)))))))
+        (when (integerp arg-spec)
+          (let* ((arg (or (nth arg-spec (combination-args node))
+                          (return-from dxify-downward-funargs)))
+                 (lvar (principal-lvar arg)))
+            (do-uses (use lvar)
+              (when (and (not (lvar-dynamic-extent arg))
+                         (ref-p use)
+                         (lambda-p (ref-leaf use))
+                         (not (leaf-dynamic-extent (functional-entry-fun (ref-leaf use))))
+                         ;; TODO: we need to do this because we don't
+                         ;; have enough smarts yet.
+                         (eq (node-home-lambda (xep-enclose (ref-leaf use)))
+                             (node-home-lambda node)))
+                (unless dynamic-extent
+                  (setq dynamic-extent (insert-dynamic-extent node)))
+                (setf (lvar-dynamic-extent arg) dynamic-extent)
+                (push arg (dynamic-extent-values dynamic-extent))))))))))
 
-;;; Starting from the potentially (declared) dynamic extent lvars
-;;; recognized during local call analysis and the declared dynamic
-;;; extent local functions recognized during IR1tran, determine if
-;;; these are actually eligible for dynamic-extent allocation. If so,
-;;; we also transitively mark the otherwise-inaccessible parts of
-;;; these values as dynamic extent.
-(defun find-dynamic-extent-lvars (component)
+;;; Check that REF delivers a value to a combination which is DX safe
+;;; or whose result is that value and ends up being discarded.
+(defun ref-good-for-dx-p (ref)
+  (let* ((lvar (ref-lvar ref))
+         (dest (when lvar (lvar-dest lvar))))
+    (and (combination-p dest)
+         (case (combination-kind dest)
+           (:known
+            (awhen (combination-fun-info dest)
+              (or (ir1-attributep (fun-info-attributes it) dx-safe)
+                  (and (not (combination-lvar dest))
+                       (awhen (fun-info-result-arg it)
+                         (eql lvar (nth it (combination-args dest))))))))
+           (:local
+            (loop for arg in (combination-args dest)
+                  for var in (lambda-vars (combination-lambda dest))
+                  do (when (eq arg lvar)
+                       (return
+                         (dolist (ref (lambda-var-refs var) t)
+                           (unless (ref-good-for-dx-p ref)
+                             (return nil)))))
+                  finally (sb-impl::unreachable)))))))
+
+;;; Recursively look for otherwise inaccessible potentially
+;;; stack-allocatable parts in the uses of LVAR. If there is one,
+;;; bound LVAR's extent by DYNAMIC-EXTENT. If LVAR already has a
+;;; different dynamic extent set, we don't do anything.
+(defun find-stack-allocatable-parts (lvar dynamic-extent)
+  (declare (type lvar lvar)
+           (type cdynamic-extent dynamic-extent))
+  (when (lvar-dynamic-extent lvar)
+    (aver (not (eq (lvar-dynamic-extent lvar) dynamic-extent)))
+    (return-from find-stack-allocatable-parts))
+  (flet ((recurse (subpart)
+           (find-stack-allocatable-parts subpart dynamic-extent)
+           (when (lvar-dynamic-extent subpart)
+             (setf (lvar-dynamic-extent lvar) dynamic-extent))))
+    (do-uses (use lvar)
+      (typecase use
+        (cast
+         (unless (cast-type-check use)
+           (recurse (cast-value use))))
+        (combination
+         (when (eq (combination-kind use) :known)
+           (let* ((info (combination-fun-info use))
+                  (stack-alloc-res (fun-info-stack-allocate-result info))
+                  (result-arg
+                    (let ((i (fun-info-result-arg info)))
+                      (and i (nth i (combination-args use))))))
+             (when result-arg
+               (recurse result-arg))
+             (when (and stack-alloc-res (funcall stack-alloc-res use))
+               (setf (lvar-dynamic-extent lvar) dynamic-extent))
+             (when (and (lvar-dynamic-extent lvar)
+                        ;; Don't propagate through &REST, for sanity.
+                        (not (eq (combination-fun-source-name use nil)
+                                 '%listify-rest-args)))
+               (dolist (arg (combination-args use))
+                 (when (and arg (not (eq result-arg arg)))
+                   (find-stack-allocatable-parts arg dynamic-extent)))))))
+        (ref
+         (let ((leaf (ref-leaf use)))
+           (typecase leaf
+             (lambda-var
+              ;; LET lambda var with no SETS.
+              (when (and (functional-kind-eq (lambda-var-home leaf) let)
+                         (not (lambda-var-sets leaf))
+                         (lexenv-contains-lambda (lambda-var-home leaf)
+                                                 (node-lexenv dynamic-extent))
+                         ;; Check the other refs are good.
+                         (dolist (ref (leaf-refs leaf) t)
+                           (unless (eq use ref)
+                             (when (not (ref-good-for-dx-p ref))
+                               (return nil)))))
+                (recurse (let-var-initial-value leaf))))
+             (clambda
+              (when (functional-kind-eq leaf external)
+                (let* ((fun (functional-entry-fun leaf))
+                       (enclose (functional-enclose fun)))
+                  (when (and (environment-closure (get-lambda-environment leaf))
+                             ;; To make sure the allocation is in the same
+                             ;; stack frame as the dynamic extent.
+                             (eq (node-home-lambda enclose)
+                                 (node-home-lambda dynamic-extent))
+                             ;; Check the other refs are good. At this
+                             ;; point, DXIFY-DOWNWARD-FUNARGS and
+                             ;; PROPAGATE-REF-DX should have marked
+                             ;; the p-lvar-ends of all good refs.
+                             (dolist (ref (leaf-refs leaf) t)
+                               (unless (eq use ref)
+                                 (multiple-value-bind (dest lvar)
+                                     (principal-lvar-end (node-lvar ref))
+                                   (declare (ignore dest))
+                                   (unless (lvar-dynamic-extent lvar)
+                                     (return nil))))))
+                    (unless (enclose-dynamic-extent enclose)
+                      (pushnew dynamic-extent
+                               (enclose-derived-dynamic-extents enclose)))
+                    (setf (lvar-dynamic-extent lvar) dynamic-extent)
+                    (setf (leaf-dynamic-extent fun) t))))))))))))
+
+;;; Find all stack allocatable values in COMPONENT, setting
+;;; appropriate dynamic extents for any lvar which may take on a stack
+;;; allocatable value. If a dynamic extent is in fact associated with
+;;; a stack allocatable value, note that fact by setting its info.
+(defun find-lvar-dynamic-extents (component)
   (declare (type component component))
   (do-blocks (block component)
-    (do-nodes (node lvar block)
+    (do-nodes (node nil block)
       (when (and (combination-p node)
                  (memq (basic-combination-kind node)
                        '(:full :unknown-keys :known)))
-        (let ((name (combination-fun-source-name node nil)))
-          (when name
-            (let ((dxable-args (fun-name-dx-args name)))
-              (when dxable-args
-                (dxify-downward-funargs node dxable-args name))))))))
+        (dxify-downward-funargs node))))
+  ;; For each dynamic extent declared variable, each value that the
+  ;; variable can take on is also dynamic extent.
   (dolist (lambda (component-lambdas component))
-    (dolist (entry (lambda-entries lambda))
-      (let* ((cleanup (entry-cleanup entry))
-             (dx-infos (cleanup-nlx-info cleanup)))
-        (when (eq (cleanup-kind cleanup) :dynamic-extent)
-          (dolist (dx-info dx-infos)
-            (let ((dx (dx-info-kind dx-info)))
-              (aver (eq cleanup (dx-info-cleanup dx-info)))
-              (labels ((mark-dx (lvar)
-                         (setf (lvar-dynamic-extent lvar) dx-info)
-                         (push lvar (dx-info-subparts dx-info))
-                         (push lvar (component-dx-lvars component))
-                         ;; Now look to see if there are otherwise
-                         ;; inaccessible parts of the value in LVAR.
-                         (do-uses (use lvar)
-                           (when (use-good-for-dx-p use cleanup dx)
-                             (etypecase use
-                               (cast (mark-dx (cast-value use)))
-                               (combination
-                                ;; Don't propagate through &REST, for
-                                ;; sanity.
-                                (unless (eq (combination-fun-source-name use nil)
-                                            '%listify-rest-args)
-                                  (dolist (arg (combination-args use))
-                                    (when (and arg
-                                               (lvar-good-for-dx-p arg cleanup dx))
-                                      (mark-dx arg)))))
-                               (ref
-                                (let ((leaf (ref-leaf use)))
-                                  (typecase leaf
-                                    (lambda-var
-                                     (mark-dx (let-var-initial-value leaf)))
-                                    (clambda
-                                     (let ((fun (functional-entry-fun leaf)))
-                                       (setf (enclose-cleanup (functional-enclose fun)) cleanup)
-                                       (setf (leaf-dynamic-extent fun) dx)))))))))))
-                (let ((lvar (dx-info-value dx-info)))
-                  ;; Check that the value hasn't been flushed somehow.
-                  (when (lvar-uses lvar)
-                    (cond ((lvar-good-for-dx-p lvar cleanup dx)
-                           (mark-dx lvar))
-                          (t
-                           (setf (lvar-dynamic-extent lvar) nil))))))))))))
-  ;; Mark closures as dynamic-extent allocatable by making the ENCLOSE
-  ;; node for the closure use an LVAR.
+    (let* ((bind (lambda-bind lambda))
+           (lexenv (node-lexenv bind))
+           dynamic-extent)
+      (dolist (var (lambda-vars lambda))
+        (when (leaf-dynamic-extent var)
+          (let ((values (mapcar #'set-value (basic-var-sets var))))
+            (when values
+              ;; This dynamic extent is over the whole environment.
+              (unless dynamic-extent
+                (setf (node-lexenv bind) (make-lexenv :default lexenv))
+                (setq dynamic-extent
+                      (with-ir1-environment-from-node bind
+                        (make-dynamic-extent)))
+                (insert-node-after bind dynamic-extent)
+                (let ((cleanup (make-cleanup :dynamic-extent dynamic-extent)))
+                  (setf (dynamic-extent-cleanup dynamic-extent) cleanup)
+                  (aver (null (lexenv-cleanup lexenv)))
+                  (setf (lexenv-cleanup lexenv) cleanup))
+                (push dynamic-extent (lambda-dynamic-extents lambda)))
+              (dolist (value values)
+                (push value (dynamic-extent-values dynamic-extent))
+                (setf (lvar-dynamic-extent value) dynamic-extent)))))))
+    (dolist (let (lambda-lets lambda))
+      (dolist (var (lambda-vars let))
+        (when (leaf-dynamic-extent var)
+          (let ((initial-value (let-var-initial-value var)))
+            ;; FIXME: This is overly pessimistic; a flushed initial
+            ;; value should not inhibit stack allocation.
+            (when initial-value
+              (let ((dynamic-extent (lvar-dynamic-extent initial-value)))
+                (dolist (set (basic-var-sets var))
+                  ;; The environments of the SET and the dynamic
+                  ;; extent must be the same to ensure the set value
+                  ;; lives on the stack long enough.
+                  (when (eq (node-environment set)
+                            (node-environment dynamic-extent))
+                    (let ((set-value (set-value set)))
+                      (unless (lvar-dynamic-extent set-value)
+                        (setf (lvar-dynamic-extent set-value) dynamic-extent)
+                        (push set-value (dynamic-extent-values dynamic-extent)))))))))))))
   (dolist (lambda (component-lambdas component))
-    (let ((fun (if (eq (lambda-kind lambda) :optional)
+    (dolist (dynamic-extent (lambda-dynamic-extents lambda))
+      (dolist (lvar (dynamic-extent-values dynamic-extent))
+        (aver (eq dynamic-extent (lvar-dynamic-extent lvar)))
+        (setf (lvar-dynamic-extent lvar) nil)
+        (find-stack-allocatable-parts lvar dynamic-extent)
+        (when (lvar-dynamic-extent lvar)
+          (unless (or (dynamic-extent-info dynamic-extent)
+                      (null (dynamic-extent-cleanup dynamic-extent)))
+            (setf (dynamic-extent-info dynamic-extent) (make-lvar)))))))
+  (dolist (lambda (component-lambdas component))
+    (let ((fun (if (functional-kind-eq lambda optional)
                    (lambda-optional-dispatch lambda)
                    lambda)))
       (when (leaf-dynamic-extent fun)
@@ -427,44 +529,19 @@
           ;; We need to have a closure environment to dynamic-extent
           ;; allocate.
           (when (and xep (environment-closure (get-lambda-environment xep)))
-            (let ((enclose (functional-enclose fun)))
-              (when (and enclose (not (node-lvar enclose)))
-                (let ((lvar (make-lvar)))
-                  (use-lvar enclose lvar)
-                  (let ((cleanup (enclose-cleanup enclose)))
-                    (if (eq enclose (cleanup-mess-up cleanup))
-                        (let ((dx-info (make-dx-info :kind 'enclose :value lvar
-                                                     :subparts (list lvar)
-                                                     :cleanup cleanup)))
-                          (setf (lvar-dynamic-extent lvar) dx-info)
-                          (dolist (dx-info (cleanup-nlx-info cleanup))
-                            (setf (lvar-dynamic-extent (dx-info-value dx-info)) nil))
-                          (setf (cleanup-nlx-info cleanup) (list dx-info)))
-                        (let* ((ref-lvar (node-lvar (first (leaf-refs xep))))
-                               (dx-info (lvar-dynamic-extent ref-lvar)))
-                          (cond (dx-info
-                                 ;; This enclose was marked DX by back
-                                 ;; propagation.
-                                 (setf (lvar-dynamic-extent lvar) dx-info)
-                                 (push lvar (dx-info-subparts dx-info)))
-                                (t
-                                 ;; This enclose was either a
-                                 ;; LET-bound anonymous lambda or
-                                 ;; inferred from a call-site to be
-                                 ;; DXable.
-                                 (let ((dx-info
-                                         (make-dx-info :kind 'enclose
-                                                       :value lvar
-                                                       :subparts (list lvar)
-                                                       :cleanup cleanup)))
-                                   (setf (lvar-dynamic-extent lvar) dx-info)
-                                   (push dx-info (cleanup-nlx-info cleanup))))))))
-                  ;; The node component of ENCLOSE may be a different
-                  ;; component for top level closure references. We
-                  ;; always compile non-top-level components before
-                  ;; top-level components, so this takes effect at the
-                  ;; right time.
-                  (push lvar (component-dx-lvars (node-component enclose)))))))))))
+            (let* ((enclose (functional-enclose fun))
+                   (dynamic-extent (enclose-dynamic-extent enclose))
+                   (derived-dynamic-extents
+                     (enclose-derived-dynamic-extents enclose)))
+              (cond (dynamic-extent
+                     (aver (null derived-dynamic-extents))
+                     (unless (dynamic-extent-info dynamic-extent)
+                       (setf (dynamic-extent-info dynamic-extent) (make-lvar))))
+                    (derived-dynamic-extents
+                     (aver (null (enclose-dynamic-extent enclose)))
+                     (let ((lvar (make-lvar)))
+                       (dolist (dynamic-extent derived-dynamic-extents)
+                         (setf (dynamic-extent-info dynamic-extent) lvar)))))))))))
   (values))
 
 ;;;; cleanup emission
@@ -510,10 +587,8 @@
              (dolist (nlx (cleanup-nlx-info cleanup))
                (code `(%lexical-exit-breakup ',nlx))))
             (:dynamic-extent
-             (dolist (dx-info (cleanup-nlx-info cleanup))
-               (when (dx-info-subparts dx-info)
-                 (code `(%cleanup-point))
-                 (return))))
+             (when (dynamic-extent-info node)
+               (code `(%cleanup-point))))
             (:restore-nsp
              (code `(%primitive set-nsp ,(ref-leaf node))))))))
     (flet ((coalesce-unbinds (code)
@@ -575,9 +650,7 @@
   (values))
 
 ;;; Mark all tail-recursive uses of function result continuations with
-;;; the corresponding TAIL-SET. Nodes whose type is NIL (i.e. don't
-;;; return) such as calls to ERROR are never annotated as TAIL-P, in
-;;; order to preserve debugging information.
+;;; the corresponding TAIL-SET.
 (defun tail-annotate (component)
   (declare (type component component))
   (dolist (fun (component-lambdas component))
@@ -586,12 +659,27 @@
         (let ((result (return-result ret)))
           (do-uses (use result)
             (when (and (immediately-used-p result use)
+                       (not (and (combination-p use)
+                                 (lvar-fun-is (combination-fun use) '(break))))
+                       (basic-combination-p use)
                        (or (not (eq (node-derived-type use) *empty-type*))
-                           (not (basic-combination-p use))
                            ;; This prevents external entry points from
                            ;; showing up in the backtrace: we always
                            ;; want tail calls inside XEPs to the
                            ;; functions they are the entry point for.
                            (eq (basic-combination-kind use) :local)))
               (setf (node-tail-p use) t)))))))
+  ;; Tail call non-returning functions if no debugging is wanted.
+  (dolist (block (block-pred (component-tail component)))
+    (let ((last (block-last block)))
+      (when (and (combination-p last)
+                 (combination-fun-info last)
+                 (policy last (= debug 0))
+                 (do-nested-cleanups (cleanup block t)
+                   (case (cleanup-kind cleanup)
+                     ((:block :tagbody)
+                      (when (entry-exits (cleanup-mess-up cleanup))
+                        (return nil)))
+                     (t (return nil)))))
+        (setf (node-tail-p last) t))))
   (values))
